# This file is part of ctrl_bps_parsl.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org) and the LSST DESC (https://www.lsstdesc.org/).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
from typing import Any, Literal, TypeVar, overload

from lsst.ctrl.bps import BpsConfig

__all__ = (
    "get_bps_config_value",
    "get_workflow_filename",
    "get_workflow_name",
    "set_parsl_logging",
)


T = TypeVar("T")


# Default provided, not required
@overload
def get_bps_config_value(
    config: BpsConfig,
    key: str,
    dataType: type[T],
    default: T,
) -> T: ...


# No default, but required
@overload
def get_bps_config_value(
    config: BpsConfig,
    key: str,
    dataType: type[T],
    default: T | None = None,
    *,
    required: Literal[True],
) -> T: ...


# No default, not required
@overload
def get_bps_config_value(
    config: BpsConfig,
    key: str,
    dataType: type[T],
    default: T | None = None,
) -> T | None: ...


def get_bps_config_value(
    config: BpsConfig,
    key: str,
    dataType: type[T],
    default: T | None = None,
    *,
    required: bool = False,
) -> T | None:
    """Get a value from the BPS configuration.

    I find this more useful than ``BpsConfig.__getitem__`` or
    ``BpsConfig.get``.

    Parameters
    ----------
    config : `BpsConfig`
        Configuration from which to retrieve value.
    key : `str`
        Key name.
    dataType : `type`
        We require that the returned value have this type.
    default : optional
        Default value to be provided if ``key`` doesn't exist in the
        ``config``. A default value of `None` means that there is no default.
    required : `bool`, optional
        If ``True``, the returned value may come from the configuration or from
        the default, but it may not be `None`.

    Returns
    -------
    value
        Value for ``key`` in the `config`` if it exists, otherwise ``default``,
        if provided.

    Raises
    ------
    KeyError
        If ``key`` is not in ``config`` and no default is provided but a value
        is ``required``.
    RuntimeError
        If the value is not set or is of the wrong type.
    """
    options: dict[str, Any] = {"expandEnvVars": True, "replaceVars": True, "required": required}
    if default is not None:
        options["default"] = default
    found, value = config.search(key, options)
    if not found and default is None:
        if required:
            raise KeyError(f"No value found for {key} and no default provided")
        return None
    if not isinstance(value, dataType):
        raise RuntimeError(f"Configuration value {key}={value} is not of type {dataType}")
    return value


def get_workflow_name(config: BpsConfig) -> str:
    """Get name of this workflow.

    The workflow name is constructed by joining the ``project`` and
    ``campaign`` (if set; otherwise ``operator``) entries in the BPS
    configuration.

    Parameters
    ----------
    config : `BpsConfig`
        BPS configuration.

    Returns
    -------
    name : `str`
        Workflow name.
    """
    project = get_bps_config_value(config, "project", str, "bps")
    campaign = get_bps_config_value(
        config, "campaign", str, get_bps_config_value(config, "operator", str, required=True)
    )
    return f"{project}.{campaign}"


def get_workflow_filename(out_prefix: str) -> str:
    """Get filename for persisting workflow.

    Parameters
    ----------
    out_prefix : `str`
        Directory which should contain workflow file.

    Returns
    -------
    filename : `str`
        Filename for persisting workflow.
    """
    return os.path.join(out_prefix, "parsl_workflow.pickle")


def set_parsl_logging(config: BpsConfig) -> int:
    """Set parsl logging levels.

    The logging level is set by the ``parsl.log_level`` entry in the BPS
    configuration.

    Parameters
    ----------
    config : `BpsConfig`
        BPS configuration.

    Returns
    -------
    level : `int`
        Logging level applied to ``parsl`` loggers.
    """
    level_name = get_bps_config_value(config, ".parsl.log_level", str, "INFO")
    if level_name not in ("CRITICAL", "DEBUG", "ERROR", "FATAL", "INFO", "WARN"):
        raise RuntimeError(f"Unrecognised parsl.log_level: {level_name}")
    level: int = getattr(logging, level_name)
    for name in logging.root.manager.loggerDict:
        if name.startswith("parsl"):
            logging.getLogger(name).setLevel(level)
    logging.getLogger("database_manager").setLevel(logging.INFO)
    return level



# This file is part of ctrl_bps_parsl.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org) and the LSST DESC (https://www.lsstdesc.org/).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Any

from parsl.executors import WorkQueueExecutor
from parsl.executors.base import ParslExecutor
from parsl.launchers import SrunLauncher
from parsl.providers import LocalProvider

try:
    from parsl.providers.base import ExecutionProvider
except ImportError:
    from parsl.providers.provider_base import ExecutionProvider  # type: ignore

from ..configuration import get_bps_config_value
from ..site import SiteConfig

if TYPE_CHECKING:
    from ..job import ParslJob

__all__ = ("LocalSrunWorkQueue", "WorkQueue")


class WorkQueue(SiteConfig):
    """Base class configuraton for `WorkQueueExecutor`.

    Subclasses must provide implementations for ``.get_executors``
    and ``.select_executor``.  In ``.get_executors``, the site-specific
    `ExecutionProvider` must be defined.

    Parameters
    ----------
    *args : `~typing.Any`
        Parameters forwarded to base class constructor.
    **kwargs : `~typing.Any`
        Keyword arguments passed to base class constructor, augmented by
        the ``add_resources`` argument.

    Notes
    -----
    The following BPS configuration parameters are recognized, overriding the
    defaults:

    - ``port`` (`int`): The port used by work_queue. Default: ``9000``.
    - ``worker_options (`str`): Extra options to pass to work_queue workers.
      A typical option specifies the memory available per worker, e.g.,
      ``"--memory=90000"``, which sets the available memory to 90 GB.
      Default: ``""``
    - ``wq_max_retries`` (`int`): The number of retries that work_queue
      will make in case of task failures.  Set to ``None`` to have work_queue
      retry forever; set to ``1`` to have retries managed by Parsl.
      Default: ``1``
    """

    def __init__(self, *args, **kwargs):
        # Have BPS-defined resource requests for each job passed to work_queue.
        kwargs["add_resources"] = True
        super().__init__(*args, **kwargs)

    def make_executor(
        self,
        label: str,
        provider: ExecutionProvider,
        *,
        port: int = 9000,
        worker_options: str = "",
        wq_max_retries: int = 1,
    ) -> ParslExecutor:
        """Return a `WorkQueueExecutor`.  The ``provider`` contains the
        site-specific configuration.

        Parameters
        ----------
        label : `str`
            Label for executor.
        provider : `ExecutionProvider`
            Parsl execution provider, e.g., `SlurmProvider`.
        port : `int`, optional
            Port used by work_queue.  Default: ``9000``.
        worker_options : `str`, optional
            Extra options to pass to work_queue workers, e.g.,
            ``"--memory=90000"``. Default: `""`.
        wq_max_retries : `int`, optional
            Number of retries for work_queue to attempt per job.  Set to
            ``None`` to have it try indefinitely; set to ``1`` to have Parsl
            control the number of retries.  Default: ``1``.
        """
        port = get_bps_config_value(self.site, "port", int, port)
        worker_options = get_bps_config_value(self.site, "worker_options", str, worker_options)
        max_retries = get_bps_config_value(self.site, "wq_max_retries", int, wq_max_retries)
        return WorkQueueExecutor(
            label=label,
            provider=provider,
            port=port,
            worker_options=worker_options,
            max_retries=max_retries,
            shared_fs=True,
            autolabel=False,
        )


class LocalSrunWorkQueue(WorkQueue):
    """Configuration for a `WorkQueueExecutor` that uses a `LocalProvider`
    to manage resources.

    This can be used directly as the site configuration within a
    multi-node allocation when Slurm is available.  For running on a
    single node, e.g., a laptop, a `SingleNodeLauncher` is used, and
    Slurm need not be available.

    The following BPS configuration parameters are recognized, overriding the
    defaults:

    - ``port`` (`int`): The port used by work_queue. Default: ``9000``.
    - ``worker_options (`str`): Extra options to pass to work_queue workers.
      A typical option specifies the memory available per worker, e.g.,
      ``"--memory=90000"``, which sets the available memory to 90 GB.
      Default: ``""``
    - ``wq_max_retries`` (`int`): The number of retries that work_queue
      will make in case of task failures.  Set to ``None`` to have work_queue
      retry forever; set to ``1`` to have retries managed by Parsl. Default:
      ``1``.
    - ``nodes_per_block`` (`int`): The number of allocated nodes.
      Default: ``1``.
    """

    def get_executors(self) -> list[ParslExecutor]:
        """Get a list of executors to be used in processing."""
        nodes_per_block = get_bps_config_value(self.site, "nodes_per_block", int, 1)
        provider_options: dict[str, Any] = {
            "nodes_per_block": nodes_per_block,
            "init_blocks": 0,
            "min_blocks": 0,
            "max_blocks": 1,
            "parallelism": 0,
            "cmd_timeout": 300,
        }
        if nodes_per_block > 1:
            provider_options["launcher"] = SrunLauncher(overrides="-K0 -k --slurmd-debug=verbose")
        provider = LocalProvider(**provider_options)
        return [self.make_executor("work_queue", provider)]

    def select_executor(self, job: "ParslJob") -> str:
        """Get the ``label`` of the executor to use to execute a job.

        Parameters
        ----------
        job : `ParslJob`
            Job to be executed.

        Returns
        -------
        label : `str`
            Label of executor to use to execute ``job``.
        """
        return "work_queue"



# This file is part of ctrl_bps_parsl.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org) and the LSST DESC (https://www.lsstdesc.org/).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any

from lsst.ctrl.bps import BaseWmsService, BaseWmsWorkflow, BpsConfig, GenericWorkflow

from .workflow import ParslWorkflow

__all__ = ("ParslService",)


class ParslService(BaseWmsService):
    """Parsl-based implementation for the WMS interface."""

    def prepare(
        self, config: BpsConfig, generic_workflow: GenericWorkflow, out_prefix: str | None = None
    ) -> BaseWmsWorkflow:
        """Convert a generic workflow to a Parsl pipeline.

        Parameters
        ----------
        config : `lsst.ctrl.bps.BpsConfig`
            Configuration of the workflow.
        generic_workflow : `lsst.ctrl.bps.generic_workflow.GenericWorkflow`
            Generic representation of a single workflow.
        out_prefix : `str` or `None`
            Prefix for WMS output files.

        Returns
        -------
        workflow : `ParslWorkflow`
            Workflow that will execute the jobs.
        """
        service_class = self.__class__.__module__ + "." + self.__class__.__name__
        if out_prefix is None:
            out_prefix = config["submitPath"]
        workflow = ParslWorkflow.from_generic_workflow(config, generic_workflow, out_prefix, service_class)
        workflow.write(out_prefix)
        return workflow

    def submit(self, workflow: BaseWmsWorkflow, **kwargs: Any):
        """Submit a single WMS workflow.

        Parameters
        ----------
        workflow : `lsst.ctrl.bps.BaseWmsWorkflow`
            Prepared WMS Workflow to submit for execution.
        **kwargs : `~typing.Any`
            Additional modifiers to the configuration.
        """
        workflow.start()
        workflow.run()

    def restart(self, out_prefix: str) -> tuple[str, str, str]:
        """Restart a workflow from the point of failure.

        Parameters
        ----------
        out_prefix : `str`
            Id for workflow to be restarted. For this service, it is the prefix
            for WMS files, also known as the ``submitPath``.

        Returns
        -------
        wms_id : `str`
            Id of the restarted workflow. If restart failed, it will be set
            to None.
        run_name : `str`
            Name of the restarted workflow. If restart failed, it will be set
            to None.
        message : `str`
            A message describing any issues encountered during the restart.
            If there were no issue, an empty string is returned.
        """
        workflow = ParslWorkflow.read(out_prefix)
        workflow.restart()
        workflow.run()
        return workflow.name, workflow.name, ""



# This file is part of ctrl_bps_parsl.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org) and the LSST DESC (https://www.lsstdesc.org/).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Any

from parsl.executors import HighThroughputExecutor
from parsl.executors.base import ParslExecutor
from parsl.launchers import MpiRunLauncher
from parsl.providers import TorqueProvider

from ..configuration import get_bps_config_value, get_workflow_name
from ..site import SiteConfig

if TYPE_CHECKING:
    from ..job import ParslJob

__all__ = ("Torque",)


Kwargs = dict[str, Any]


class Torque(SiteConfig):
    """Configuration for generic Torque cluster.

    This can be used directly as the site configuration for a Torque cluster by
    setting the BPS config, e.g.:

    .. code-block:: yaml

        computeSite: torque
        site:
          torque:
            class: lsst.ctrl.bps.parsl.sites.Torque
            nodes: 4
            tasks_per_node: 20
            walltime: "00:59:00"  # Note: always quote walltime in YAML

    Alternatively, it can be used as a base class for Torque cluster
    configurations.

    The following BPS configuration parameters are recognised (and required
    unless there is a default mentioned here, or provided by a subclass):

    - ``queue`` (`int`): Queue for the Torque job.
    - ``nodes`` (`int`): number of nodes for each Torque job.
    - ``tasks_per_node`` (`int`): number of cores per node for each Torque job;
      by default we use all cores on the node.
    - ``walltime`` (`str`): time limit for each Torque job.
    - ``scheduler_options`` (`str`): text to prepend to the Torque submission
      script (each line usually starting with ``#PBS``).
    """

    def make_executor(
        self,
        label: str,
        *,
        queue: str | None = None,
        nodes: int | None = None,
        tasks_per_node: int | None = None,
        walltime: str | None = None,
        mem_per_worker: float | None = None,
        scheduler_options: str | None = None,
        worker_init: str | None = None,
        provider_options: Kwargs | None = None,
        executor_options: Kwargs | None = None,
    ) -> ParslExecutor:
        """Return an executor for running on a Torque cluster.

        Parameters
        ----------
        label : `str`
            Label for executor.
        queue : `str`, optional
            Queue for the Torque job.
        nodes : `int`, optional
            Default number of nodes for each Torque job.
        tasks_per_node : `int`, optional
            Default number of cores per node for each Torque job.
        walltime : `str`, optional
            Default time limit for each Torque job.
        mem_per_worker : `float`, optional
            Minimum memory per worker (GB), limited by the executor.
        worker_init : `str`, optional
            Environment initiation command
        scheduler_options : `str`, optional
            ``#SBATCH`` directives to prepend to the Torque submission script.
        provider_options : `dict`, optional
            Additional arguments for `TorqueProvider` constructor.
        executor_options : `dict`, optional
            Additional arguments for `HighThroughputExecutor` constructor.

        Returns
        -------
        executor : `HighThroughputExecutor`
            Executor for Torque jobs.
        """
        nodes = get_bps_config_value(self.site, "nodes", int, nodes, required=True)
        walltime = get_bps_config_value(self.site, "walltime", str, walltime, required=True)
        queue = get_bps_config_value(self.site, "queue", str, queue)
        tasks_per_node = get_bps_config_value(self.site, "tasks_per_node", int, tasks_per_node)
        worker_init = get_bps_config_value(self.site, "worker_init", str, "")
        scheduler_options = get_bps_config_value(self.site, "scheduler_options", str, scheduler_options)

        if tasks_per_node is None:
            tasks_per_node = 1

        job_name = get_workflow_name(self.config)

        if scheduler_options is None:
            scheduler_options = ""
        else:
            scheduler_options += "\n"
        scheduler_options += f"#PBS -N {job_name}\n"
        if queue:
            scheduler_options += f"#PBS -q {queue}\n"

        if worker_init is None:
            worker_init = ""

        launcher = PbsMpiRunLauncher(overrides=f"--map-by core:{tasks_per_node}")

        return HighThroughputExecutor(
            label,
            provider=PbsTorqueProvider(
                nodes_per_block=nodes,
                tasks_per_node=tasks_per_node,
                queue=queue,
                walltime=walltime,
                scheduler_options=scheduler_options,
                worker_init=worker_init,
                launcher=launcher,
                **(provider_options or {}),
            ),
            max_workers_per_node=1,
            mem_per_worker=mem_per_worker,
            address=self.get_address(),
            **(executor_options or {}),
        )

    def get_executors(self) -> list[ParslExecutor]:
        """Get a list of executors to be used in processing.

        Each executor should have a unique ``label``.
        """
        return [self.make_executor("torque")]

    def select_executor(self, job: "ParslJob") -> str:
        """Get the ``label`` of the executor to use to execute a job.

        Parameters
        ----------
        job : `ParslJob`
            Job to be executed.

        Returns
        -------
        label : `str`
            Label of executor to use to execute ``job``.
        """
        return "torque"


class PbsTorqueProvider(TorqueProvider):
    """Torque Execution Provider

    This provider uses qsub to submit, qstat for status, and qdel to cancel
    jobs. The qsub script to be used is created from a template file in this
    same module.

    This subclass allows the ``tasks_per_node`` to be set at construction time
    instead of at submission time.
    """

    def __init__(self, *args, tasks_per_node: int = 1, **kwargs):
        super().__init__(*args, **kwargs)
        self.tasks_per_node = tasks_per_node

    def submit(self, command, tasks_per_node, job_name="parsl.torque"):
        """Submit the command onto an Local Resource Manager job.

        This function returns an ID that corresponds to the task that was just
        submitted.

        The ``tasks_per_node`` parameter is ignored in this provider, as it is
        set at construction time.

        Parameters
        ----------
        command : `str`
            Command-line invocation to be made on the remote side.
        tasks_per_node : `int`
            Number of tasks to be launched per node. This is ignored in this
            provider.
        job_name : `str`:
            Name for job, must be unique.

        Returns
        -------
            None: At capacity, cannot provision more
            job_id (string): Identifier for the job

        """
        return super().submit(
            command=command,
            tasks_per_node=self.tasks_per_node,
            job_name=job_name,
        )


class PbsMpiRunLauncher(MpiRunLauncher):
    """Worker launcher that wraps the user's command with the framework to
    launch multiple command invocations via ``mpirun``.

    This wrapper sets the bash env variable ``CORES`` to the number of cores on
    the machine.

    This launcher makes the following assumptions:
    - mpirun is installed and can be located in ``$PATH``
    - The provider makes available the ``$PBS_NODEFILE`` environment variable
    """

    def __init__(
        self,
        debug: bool = True,
        bash_location: str = "/bin/bash",
        overrides: str = "",
    ):
        super().__init__(debug=debug, bash_location=bash_location, overrides=overrides)

    def __call__(self, command: str, tasks_per_node: int, nodes_per_block: int) -> str:
        """Wrap the user's command with mpirun invocation"""
        worker_count = nodes_per_block * tasks_per_node
        debug_num = int(self.debug)

        return f"""set -e
export CORES=$(getconf _NPROCESSORS_ONLN)
[[ "{debug_num}" == "1" ]] && echo "Found cores : $CORES"
WORKERCOUNT={worker_count}

cat << MPIRUN_EOF > cmd_$JOBNAME.sh
{command}
MPIRUN_EOF
chmod u+x cmd_$JOBNAME.sh

mpirun -np $WORKERCOUNT {self.overrides} {self.bash_location} cmd_$JOBNAME.sh

[[ "{debug_num}" == "1" ]] && echo "All workers done"
"""



# This file is part of ctrl_bps_parsl.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org) and the LSST DESC (https://www.lsstdesc.org/).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any

from parsl.executors.base import ParslExecutor
from parsl.launchers import SrunLauncher

from .slurm import TripleSlurm

Kwargs = dict[str, Any]

__all__ = ("CoriKnl",)


class CoriKnl(TripleSlurm):
    """Configuration for running jobs on the NERSC Cori-KNL cluster.

    Notes
    -----
    The following BPS configuration parameters are recognised, overriding the
    defaults:

    - ``nodes`` (`int`): number of nodes for each Slurm job.
    - ``cores_per_node`` (`int`): number of cores per node for each Slurm job;
      by default we use all cores on the node.
    - ``walltime`` (`str`): time limit for each Slurm job; setting this would
      override each of the ``small_walltime``, ``medium_walltime`` and
      ``large_walltime`` values.
    - ``mem_per_node`` (`float`): memory per node for each Slurm job; by
      default we use whatever Slurm gives us.
    - ``qos`` (`str`): quality of service to request for each Slurm job; by
      default we use whatever Slurm gives us.
    - ``small_memory`` (`float`): memory per worker for each 'small' Slurm job.
    - ``medium_memory`` (`float`): memory per worker for each 'small' Slurm
      job.
    - ``large_memory`` (`float`): memory per worker for each 'large' Slurm job.
    - ``small_walltime`` (`str`): time limit for each 'small' Slurm job.
    - ``medium_walltime`` (`str`): time limit for each 'small' Slurm job.
    - ``large_walltime`` (`str`): time limit for each 'large' Slurm job.
    """

    def get_executors(
        self,
        small_options: Kwargs | None = None,
        medium_options: Kwargs | None = None,
        large_options: Kwargs | None = None,
        **common_options,
    ) -> list[ParslExecutor]:
        """Get a list of executors to be used in processing.

        We create three executors, with different walltime and memory per
        worker.

        Parameters
        ----------
        small_options : `dict`
            Options for ``make_executor`` for small executor.
        medium_options : `dict`
            Options for ``make_executor`` for medium executor.
        large_options : `dict`
            Options for ``make_executor`` for large executor.
        **common_options : Any
            Common options for ``make_executor`` for each of the executors.
        """
        scheduler_options = "\n".join(
            (
                "#SBATCH --module=cvmfs",
                "#SBATCH --licenses=cvmfs",
                "#SBATCH --time-min=2:00:00",
            )
        )
        provider_options = {
            "exclusive": True,
            "init_blocks": 0,
            "min_blocks": 0,
            "max_blocks": 1,
            "parallelism": 0,
            "launcher": SrunLauncher(overrides="-K0 -k --slurmd-debug=verbose"),
            "cmd_timeout": 300,
        }
        executor_options = {"worker_debug": False, "heartbeat_period": 60, "heartbeat_threshold": 180}

        return super().get_executors(
            nodes=1,
            constraint="knl",
            scheduler_options=scheduler_options,
            provider_options=provider_options,
            executor_options=executor_options,
            **common_options,
        )



# This file is part of ctrl_bps_parsl.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org) and the LSST DESC (https://www.lsstdesc.org/).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
import pickle
from collections.abc import Iterable, Mapping

import parsl
import parsl.config
from parsl.app.app import bash_app
from parsl.app.bash import BashApp
from parsl.app.futures import Future

from lsst.ctrl.bps import BaseWmsWorkflow, BpsConfig, GenericWorkflow, GenericWorkflowJob

from .configuration import get_workflow_filename, set_parsl_logging
from .job import ParslJob, get_file_paths
from .site import SiteConfig

__all__ = ("ParslWorkflow", "get_parsl_config")

_log = logging.getLogger("lsst.ctrl.bps.parsl")


def get_parsl_config(config: BpsConfig) -> parsl.config.Config:
    """Construct parsl configuration from BPS configuration.

    For details on the site configuration, see `SiteConfig`. For details on the
    monitor configuration, see ``get_parsl_monitor``.

    `SiteConfig` provides an implementation of the method ``get_parsl_config``
    which returns a Parsl configuration with sensible defaults. Subclasses
    of `SiteConfig` can overwrite that method to configure Parsl in a
    way specific to the site's configuration.

    Parameters
    ----------
    config : `BpsConfig`
        BPS configuration.

    Returns
    -------
    parsl_config : `parsl.config.Config`
        Parsl configuration.
    """
    site = SiteConfig.from_config(config)
    return site.get_parsl_config()


class ParslWorkflow(BaseWmsWorkflow):
    """Parsl-based workflow object to manage execution of workflow.

    Parameters
    ----------
    name : `str`
        Unique name of workflow.
    config : `lsst.ctrl.bps.BpsConfig`
        Generic workflow config.
    path : `str`
        Path prefix for workflow output files.
    jobs : `dict` mapping `str` to `ParslJob`
        Jobs to be executed.
    parents : `dict` mapping `str` to iterable of `str`
        Dependency tree. Keywords are job names, and values are a list of job
        names that must be executed before the keyword job name can be
        executed.
    endpoints : iterable of `str`
        Endpoints of the dependency tree. These jobs (specified by name) have
        no children.
    final : `ParslJob`, optional
        Final job to be done, e.g., to merge the execution butler. This is done
        locally.
    """

    def __init__(
        self,
        name: str,
        config: BpsConfig,
        path: str,
        jobs: dict[str, ParslJob],
        parents: Mapping[str, Iterable[str]],
        endpoints: Iterable[str],
        final: ParslJob | None = None,
    ):
        super().__init__(name, config)

        self.path = path
        self.bps_config = config
        self.parsl_config = get_parsl_config(config)
        self.site = SiteConfig.from_config(config)
        self.dfk: parsl.DataFlowKernel | None = None  # type: ignore
        self.command_prefix = self.site.get_command_prefix()

        # these are function decorators
        self.apps: dict[str, BashApp] = {
            ex.label: bash_app(  # type: ignore
                executors=[ex.label], cache=True, ignore_for_cache=["stderr", "stdout"]
            )
            for ex in self.parsl_config.executors
        }

        self.jobs = jobs
        self.parents = parents
        self.endpoints = endpoints
        self.final = final

    def __reduce__(self):
        """Recipe for pickle"""
        return type(self), (
            self.name,
            self.bps_config,
            self.path,
            self.jobs,
            self.parents,
            self.endpoints,
            self.final,
        )

    @classmethod
    def from_generic_workflow(
        cls, config: BpsConfig, generic_workflow: GenericWorkflow, out_prefix: str, service_class: str
    ) -> BaseWmsWorkflow:
        """Create a ParslWorkflow object from a BPS GenericWorkflow.

        Parameters
        ----------
        config : `BpsConfig`
            Configuration of the workflow.
        generic_workflow : `lsst.ctrl.bps.generic_workflow.GenericWorkflow`
            Generic representation of a single workflow.
        out_prefix : `str`
            Prefix for workflow output files.
        service_class : `str`
            Full module name of WMS service class that created this workflow.

        Returns
        -------
        self : `ParslWorkflow`
            Constructed workflow.
        """
        # Generate list of jobs
        jobs: dict[str, ParslJob] = {}
        for job_name in generic_workflow:
            generic_job = generic_workflow.get_job(job_name)
            assert generic_job.name not in jobs
            jobs[job_name] = ParslJob(generic_job, config, get_file_paths(generic_workflow, job_name))

        parents = {name: set(generic_workflow.predecessors(name)) for name in jobs}
        endpoints = [name for name in jobs if generic_workflow.out_degree(name) == 0]

        # Add final job: execution butler merge
        job = generic_workflow.get_final()
        final: ParslJob | None = None
        if job is not None:
            assert isinstance(job, GenericWorkflowJob)
            final = ParslJob(job, config, get_file_paths(generic_workflow, job.name))

        return cls(generic_workflow.name, config, out_prefix, jobs, parents, endpoints, final)

    def write(self, out_prefix: str):
        """Write workflow state.

        This, in combination with the parsl checkpoint files, can be used to
        restart a workflow that was interrupted.

        Parameters
        ----------
        out_prefix : `str`
            Root directory to be used for WMS workflow inputs and outputs
            as well as internal WMS files.
        """
        filename = get_workflow_filename(out_prefix)
        _log.info("Writing workflow with ID=%s", out_prefix)
        with open(filename, "wb") as fd:
            pickle.dump(self, fd)

    @classmethod
    def read(cls, out_prefix: str) -> "ParslWorkflow":
        """Construct from the saved workflow state.

        Parameters
        ----------
        out_prefix : `str`
            Root directory to be used for WMS workflow inputs and outputs
            as well as internal WMS files.

        Returns
        -------
        self : `ParslWorkflow`
            Constructed workflow.
        """
        filename = get_workflow_filename(out_prefix)
        with open(filename, "rb") as fd:
            self = pickle.load(fd)
        assert isinstance(self, cls)
        return self

    def run(self, block: bool = True) -> list[Future | None]:
        """Run the workflow.

        Parameters
        ----------
        block : `bool`, optional
            Block returning from this method until the workflow is complete? If
            `False`, jobs may still be running when this returns, and it is the
            user's responsibility to call the ``finalize_jobs`` and
            ``shutdown`` methods when they are complete.

        Returns
        -------
        futures : `list` of `Future`
            `Future` objects linked to the execution of the endpoint jobs.
        """
        futures = [self.execute(name) for name in self.endpoints]
        if block:
            # Calling .exception() for each future blocks returning
            # from this method until all the jobs have executed or
            # raised an error.  This is needed for running in a
            # non-interactive python process that would otherwise end
            # before the futures resolve.
            for ff in futures:
                if ff is not None:
                    ff.exception()
            self.shutdown()
            self.finalize_jobs()
        return futures

    def execute(self, name: str) -> Future | None:
        """Execute a job.

        Parameters
        ----------
        name : `str`
            Name of job to execute.

        Returns
        -------
        future : `Future` or `None`
            A `Future` object linked to the execution of the job, or `None` if
            the job is being reserved to run locally.
        """
        if name in ("pipetaskInit", "mergeExecutionButler"):
            # These get done outside of parsl
            return None
        job = self.jobs[name]
        inputs = [self.execute(parent) for parent in self.parents[name]]
        executors = self.parsl_config.executors
        if len(executors) > 1:
            label = self.site.select_executor(job)
        else:
            label = executors[0].label
        return job.get_future(
            self.apps[label],
            [ff for ff in inputs if ff is not None],
            self.command_prefix,
            self.site.add_resources,
        )

    def load_dfk(self):
        """Load data frame kernel.

        This starts parsl.
        """
        if self.dfk is not None:
            raise RuntimeError("Workflow has already started.")
        set_parsl_logging(self.bps_config)
        self.dfk = parsl.load(self.parsl_config)

    def start(self):
        """Start the workflow."""
        self.initialize_jobs()
        self.load_dfk()

    def restart(self):
        """Restart the workflow after interruption."""
        self.parsl_config.checkpoint_files = parsl.utils.get_last_checkpoint()
        self.load_dfk()

    def shutdown(self):
        """Shut down the workflow.

        This stops parsl.
        """
        if self.dfk is None:
            raise RuntimeError("Workflow not started.")
        self.dfk.cleanup()
        self.dfk = None
        parsl.DataFlowKernelLoader.clear()

    def initialize_jobs(self):
        """Run initial jobs.

        These jobs are run locally before any other jobs are submitted to
        parsl.

        This is used to set up the butler.
        """
        job = self.jobs.get("pipetaskInit", None)
        if job is not None:
            os.makedirs(os.path.join(self.path, "logs"))
            job.run_local()

    def finalize_jobs(self):
        """Run final jobs.

        These jobs are run locally after all other jobs are complete.

        This is used to merge the execution butler.
        """
        if self.final is not None and not self.final.done:
            self.final.run_local()



# This file is part of ctrl_bps_parsl.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org) and the LSST DESC (https://www.lsstdesc.org/).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from types import ModuleType
from typing import TYPE_CHECKING

import parsl.config
from parsl.addresses import address_by_hostname
from parsl.executors.base import ParslExecutor
from parsl.monitoring import MonitoringHub

from lsst.ctrl.bps import BpsConfig
from lsst.utils import doImport

from .configuration import get_bps_config_value, get_workflow_name
from .environment import export_environment

if TYPE_CHECKING:
    from .job import ParslJob

__all__ = ("SiteConfig",)


class SiteConfig(ABC):
    """Base class for site configuration.

    Subclasses need to override at least the ``get_executors`` and
    ``select_executor`` methods.

    Parameters
    ----------
    config : `BpsConfig`
        BPS configuration.
    add_resources : `bool`
        Add resource specification when submitting the job? This is only
        appropriate for the ``WorkQueue`` executor; other executors will
        raise an exception.
    """

    def __init__(self, config: BpsConfig, add_resources: bool = False):
        self.config = config
        self.site = self.get_site_subconfig(config)
        self.add_resources = add_resources

    @staticmethod
    def get_site_subconfig(config: BpsConfig) -> BpsConfig:
        """Get BPS configuration for the site of interest.

        We return the BPS sub-configuration for the site indicated by the
        ``computeSite`` value, which is ``site.<computeSite>``.

        Parameters
        ----------
        config : `BpsConfig`
            BPS configuration.

        Returns
        -------
        site : `BpsConfig`
            Site sub-configuration.
        """
        computeSite = get_bps_config_value(config, "computeSite", str, required=True)
        return get_bps_config_value(config, f".site.{computeSite}", BpsConfig, required=True)

    @classmethod
    def from_config(cls, config: BpsConfig) -> "SiteConfig":
        """Get the site configuration nominated in the BPS config.

        The ``computeSite`` (`str`) value in the BPS configuration is used to
        select a site configuration. The site configuration class to use is
        specified by the BPS configuration as ``site.<computeSite>.class``
        (`str`), which should be the fully-qualified name of a python class
        that inherits from `SiteConfig`.

        Parameters
        ----------
        config : `BpsConfig`
            BPS configuration.

        Returns
        -------
        site_config : subclass of `SiteConfig`
            Site configuration.
        """
        site = cls.get_site_subconfig(config)
        name = get_bps_config_value(site, "class", str, required=True)
        site_config = doImport(name)
        if isinstance(site_config, ModuleType) or not issubclass(site_config, SiteConfig):
            raise RuntimeError(f"Site class={name} is not a SiteConfig subclass")
        return site_config(config)

    @abstractmethod
    def get_executors(self) -> list[ParslExecutor]:
        """Get a list of executors to be used in processing.

        Each executor should have a unique ``label``.
        """
        raise NotImplementedError("Subclasses must define")

    @abstractmethod
    def select_executor(self, job: "ParslJob") -> str:
        """Get the ``label`` of the executor to use to execute a job.

        Parameters
        ----------
        job : `ParslJob`
            Job to be executed.

        Returns
        -------
        label : `str`
            Label of executor to use to execute ``job``.
        """
        raise NotImplementedError("Subclasses must define")

    def get_address(self) -> str:
        """Return the IP address of the machine hosting the driver/submission.

        This address should be accessible from the workers. This should
        generally by the return value of one of the functions in
        ``parsl.addresses``.

        This is used by the default implementation of ``get_monitor``, but will
        generally be used by ``get_executors`` too.

        This default implementation gets the address from the hostname, but
        that will not work if the workers don't access the driver/submission
        node by that address.
        """
        return address_by_hostname()

    def get_command_prefix(self) -> str:
        """Return command(s) to add before each job command.

        These may be used to configure the environment for the job.

        This default implementation respects the BPS configuration elements:

        - ``site.<computeSite>.commandPrefix`` (`str`): command(s) to use as a
          prefix to executing a job command on a worker.
        - ``site.<computeSite>.environment`` (`bool`): add bash commands that
          replicate the environment on the driver/submit machine?
        """
        prefix = get_bps_config_value(self.site, "commandPrefix", str, "")
        if get_bps_config_value(self.site, "environment", bool, False):
            prefix += "\n" + export_environment()
        return prefix

    def get_monitor(self) -> MonitoringHub | None:
        """Get parsl monitor.

        The parsl monitor provides a database that tracks the progress of the
        workflow and the use of resources on the workers.

        This implementation respects the BPS configuration elements:

        - ``site.<computeSite>.monitorEnable`` (`bool`): enable monitor?
        - ``site.<computeSite>.monitorInterval`` (`float`): time interval (sec)
          between logging of resource usage.
        - ``site.<computeSite>.monitorFilename`` (`str`): name of file to use
          for the monitor sqlite database.

        Returns
        -------
        monitor : `MonitoringHub` or `None`
            Parsl monitor, or `None` for no monitor.
        """
        if not get_bps_config_value(self.site, "monitorEnable", bool, False):
            return None
        return MonitoringHub(
            workflow_name=get_workflow_name(self.config),
            hub_address=self.get_address(),
            resource_monitoring_interval=get_bps_config_value(self.site, "monitorInterval", float, 30.0),
            logging_endpoint="sqlite:///"
            + get_bps_config_value(self.site, "monitorFilename", str, "monitor.sqlite"),
        )

    def get_parsl_config(self) -> parsl.config.Config:
        """Get Parsl configuration for this site.

        Subclasses can overwrite this method to build a more specific Parsl
        configuration, if required.

        The retries are set from the ``site.<computeSite>.retries`` value
        found in the BPS configuration file.

        Returns
        -------
        config : `parsl.config.Config`
            The configuration to be used for Parsl.
        """
        executors = self.get_executors()
        monitor = self.get_monitor()
        retries = get_bps_config_value(self.site, "retries", int, 1)
        # Path to Parsl run directory. The default set by Parsl is
        # 'runinfo' which is not explicit enough for end users given that
        # we are using BPS + Parsl + Slurm to execute a workflow.
        run_dir = get_bps_config_value(self.site, "run_dir", str, "runinfo")
        return parsl.config.Config(
            executors=executors,
            monitoring=monitor,
            retries=retries,
            run_dir=run_dir,
            checkpoint_mode="task_exit",
        )



# This file is part of ctrl_bps_parsl.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org) and the LSST DESC (https://www.lsstdesc.org/).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING

from parsl.addresses import address_by_interface
from parsl.executors.base import ParslExecutor
from parsl.launchers import SrunLauncher
from psutil import net_if_addrs

from ..configuration import get_bps_config_value
from ..environment import export_environment
from .slurm import Slurm

if TYPE_CHECKING:
    from ..job import ParslJob

__all__ = ("Tiger",)


class Tiger(Slurm):
    """Configuration for running jobs on Princeton's Tiger cluster.

    The following BPS configuration parameters are recognised, overriding the
    defaults:

    - ``nodes`` (`int`): number of nodes for each Slurm job.
    - ``cores_per_node`` (`int`): number of cores per node for each Slurm job.
    - ``walltime`` (`str`): time limit for each Slurm job.
    - ``mem_per_node`` (`int`): memory per node (GB) for each Slurm job.
    - ``max_blocks`` (`int`): maximum number of blocks (Slurm jobs) to use.
    - ``cmd_timeout`` (`int`): timeout (seconds) to wait for a scheduler.
    - ``singleton`` (`bool`): allow only one job to run at a time; by default
      ``True``.

    When running on the Tiger cluster, you should operate on the
    ``/scratch/gpfs`` filesystem, rather than ``/projects`` or ``/tigress``;
    the latter are not even mounted on the cluster nodes any more.
    """

    def get_executors(self) -> list[ParslExecutor]:
        """Get a list of executors to be used in processing.

        Each executor should have a unique ``label``.

        The walltime default here is set so we get into the tiger-vshort QoS,
        which will hopefully reduce the wait for us to get a node. Then, we
        have one Slurm job running at a time (singleton) while another saves a
        spot in line (max_blocks=2). We hope that this will allow us to run
        almost continually until the workflow is done.

        We set the cmd_timeout value to 300 seconds to help avoid
        TimeoutExpired errors when commands are slow to return (often due to
        system contention).
        """
        max_blocks = get_bps_config_value(self.site, "max_blocks", int, 2)
        cmd_timeout = get_bps_config_value(self.site, "cmd_timeout", int, 300)
        return [
            self.make_executor(
                "tiger",
                nodes=4,
                cores_per_node=112,
                walltime="05:00:00",  # Ensures we get into qos=tiger-vshort, which cuts off at 5h
                mem_per_node=980,  # Ensures all nodes are available, reserving a little for OS services
                singleton=True,
                provider_options={
                    "init_blocks": 1,
                    "min_blocks": 1,
                    "max_blocks": max_blocks,
                    "parallelism": 1.0,
                    "worker_init": export_environment(),
                    "launcher": SrunLauncher(overrides="-K0 -k"),
                    "cmd_timeout": cmd_timeout,
                },
            )
        ]

    def select_executor(self, job: "ParslJob") -> str:
        """Get the ``label`` of the executor to use to execute a job.

        Parameters
        ----------
        job : `ParslJob`
            Job to be executed.

        Returns
        -------
        label : `str`
            Label of executor to use to execute ``job``.
        """
        return "tiger"

    def get_address(self) -> str:
        """Return the IP address of the machine hosting the driver/submission.

        This host machine address should be accessible from the workers and
        should generally be the return value of one of the functions in
        ``parsl.addresses``.

        This is used by the default implementation of ``get_monitor``, but will
        generally be used by ``get_executors`` too.

        This implementation gets the address from the Infiniband network
        interface, because the cluster nodes can't connect to the head node
        through the regular internet.
        """
        net_interfaces = [interface for interface in net_if_addrs().keys() if interface[:2] in ["ib", "op"]]
        if net_interfaces:
            return address_by_interface(net_interfaces[0])
        raise RuntimeError("No Infiniband network interface found.")



# This file is part of ctrl_bps_parsl.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org) and the LSST DESC (https://www.lsstdesc.org/).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING

from parsl.executors.base import ParslExecutor
from parsl.launchers import SrunLauncher

from ..configuration import get_bps_config_value
from ..environment import export_environment
from .slurm import Slurm

if TYPE_CHECKING:
    from ..job import ParslJob

__all__ = ("Sdf",)


class Sdf(Slurm):
    """Slurm-based configuration for running jobs on SLAC's Shared Data
    Facility cluster using the rubin partition.

    The following BPS configuration parameters are recognised, overriding the
    defaults:

    - ``nodes`` (`int`): number of nodes for each Slurm job.
    - ``cores_per_node`` (`int`): number of cores per node for each Slurm job.
    - ``walltime`` (`str`): time limit for each Slurm job.
    - ``mem_per_node`` (`int`): memory per node (GB) for each Slurm job.
    - ``max_blocks`` (`int`): maximum number of blocks (Slurm jobs) to use.
    - ``singleton`` (`bool`): allow only one job to run at a time; by default
      ``True``.
    """

    def get_executors(self) -> list[ParslExecutor]:
        """Get a list of executors to be used in processing.

        Each executor should have a unique ``label``.

        We have one Slurm job running at a time (singleton) while
        another saves a spot in line (max_blocks=2). We hope that this
        allow us to run almost continually until the workflow is done.
        """
        max_blocks = get_bps_config_value(self.site, "max_blocks", int, 2)
        return [
            self.make_executor(
                "sdf",
                nodes=1,
                cores_per_node=100,
                walltime="02:00:00",
                mem_per_node=400,
                qos="normal",
                singleton=True,
                scheduler_options="#SBATCH --partition=rubin",
                provider_options={
                    "init_blocks": 1,
                    "min_blocks": 1,
                    "max_blocks": max_blocks,
                    "parallelism": 1.0,
                    "worker_init": export_environment(),
                    "launcher": SrunLauncher(overrides="-K0 -k --slurmd-debug=verbose"),
                },
            )
        ]

    def select_executor(self, job: "ParslJob") -> str:
        """Get the ``label`` of the executor to use to execute a job.

        Parameters
        ----------
        job : `ParslJob`
            Job to be executed.

        Returns
        -------
        label : `str`
            Label of executor to use to execute ``job``.
        """
        return "sdf"



import copy
import platform
from typing import TYPE_CHECKING, Any

import parsl.config
from parsl.executors import HighThroughputExecutor
from parsl.executors.base import ParslExecutor
from parsl.providers import SlurmProvider

from ..configuration import get_bps_config_value
from ..site import SiteConfig

if TYPE_CHECKING:
    from .job import ParslJob

__all__ = ("Ccin2p3",)

Kwargs = dict[str, Any]


class Ccin2p3(SiteConfig):
    """Configuration for executing Parsl jobs in CC-IN2P3 Slurm batch farm.

    This class provides four job slot sizes each with its specific
    requirements, in particular in terms of memory. Those slot sizes are named
    "small", "medium", "large" and "xlarge".

    Sensible default values for those requirements are provided for each
    job slot but you can overwrite those defaults either in the
    the BPS submission file or in a site configuration file that you
    include in your BPS submission file.

    If you don't need to modify the default requirements for the job slot
    sizes, use the site specification below in your BPS configuration
    file:

    .. code-block:: yaml

        wmsServiceClass: lsst.ctrl.bps.parsl.ParslService
        computeSite: ccin2p3

        site:
          ccin2p3:
            class: lsst.ctrl.bps.parsl.sites.ccin2p3.Ccin2p3

    If you do need to modify those defaults, you can overwrite them for
    all job slots or for specific each job slots. Requirements specified
    for a job slot take priority over those specified for all job slots
    at the level of entry '.site.ccin2p3:'.

    This is an example of how to overwrite selected requirements in your BPS
    submission file:

    .. code-block:: yaml

        wmsServiceClass: lsst.ctrl.bps.parsl.ParslService
        computeSite: ccin2p3

        site:
          ccin2p3:
            class: lsst.ctrl.bps.parsl.sites.ccin2p3.Ccin2p3
            walltime: "72:00:00"
            scheduler_options:
              - "--licenses=sps"
              - "--qos=normal"
            small:
              memory: 6
              partition: "flash"
            medium:
              memory: 10
              partition: "lsst,htc"
            large:
              memory: 80
            xlarge:
              memory: 180
              partition: "lsst"
              scheduler_options:
                - "--constraint=el7"
                - "--licenses=my_product"
                - "--reservation=my_reservation"

    At the level of entry 'site.ccin2p3:' in the BPS submission file, the
    following configuration parameters are accepted, which apply to all slot
    sizes:

    - `partition` (`str`): name of the one or more configured partitions. If
       more than one, separate them with comma (',').
       (Default: "lsst,htc")
    - `walltime` (`str`): walltime to require for the job (Default: "72:00:00")
    - `scheduler_options` (`list` [`str`] ): scheduler options to send to Slurm
       for scheduling purposes.
       (Default: "--licenses=sps")

    In addition, as shown in the previous example, for each job slot (i.e.
    "small", "medium", etc.) you can specify the requirements above as well as
    the following:

    - `max_blocks` (`int`): maximum number of Slurm jobs that your workflow can
       simultaneously use.
    - `memory` (`int`): required amount of memory for each job, in Gigabytes.
       (Defaults: 4 for "small", 10 for "medium", 50 fo "large" and
       150 for "xlarge").

    Parameters
    ----------
    *args : optional
        Arguments to initialize the super-class.
    **kwargs : optional
        Keyword arguments to initialize the super-class.

    Returns
    -------
    Ccin2p3 : `SiteConfig`
        Concrete instance of a `SiteConfig` specific for the CC-IN2P3 Slurm
        farm.
    """

    DEFAULT_ACCOUNT: str = "lsst"
    DEFAULT_WALLTIME: str = "72:00:00"
    DEFAULT_SCHEDULER_OPTIONS: list[str] = [
        "--licenses=sps",
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._account = get_bps_config_value(self.site, ".account", str, self.DEFAULT_ACCOUNT)
        self._scheduler_options = get_bps_config_value(
            self.site, ".scheduler_options", list, self.DEFAULT_SCHEDULER_OPTIONS
        )
        self._slot_size = {
            "small": {
                "memory": get_bps_config_value(self.site, ".small.memory", int, 4),
                "walltime": self._get_walltime_for_slot("small"),
                "partition": self._get_partition_for_slot("small"),
                "max_blocks": get_bps_config_value(self.site, ".small.max_blocks", int, 3_000),
                "scheduler_options": get_bps_config_value(self.site, ".small.scheduler_options", list, []),
            },
            "medium": {
                "memory": get_bps_config_value(self.site, ".medium.memory", int, 10),
                "walltime": self._get_walltime_for_slot("medium"),
                "partition": self._get_partition_for_slot("medium"),
                "max_blocks": get_bps_config_value(self.site, ".medium.max_blocks", int, 1_000),
                "scheduler_options": get_bps_config_value(self.site, ".medium.scheduler_options", list, []),
            },
            "large": {
                "memory": get_bps_config_value(self.site, ".large.memory", int, 50),
                "walltime": self._get_walltime_for_slot("large"),
                "partition": self._get_partition_for_slot("large"),
                "max_blocks": get_bps_config_value(self.site, ".large.max_blocks", int, 100),
                "scheduler_options": get_bps_config_value(self.site, ".large.scheduler_options", list, []),
            },
            "xlarge": {
                "memory": get_bps_config_value(self.site, ".xlarge.memory", int, 150),
                "walltime": self._get_walltime_for_slot("xlarge"),
                "partition": self._get_partition_for_slot("xlarge"),
                "max_blocks": get_bps_config_value(self.site, ".xlarge.max_blocks", int, 10),
                "scheduler_options": get_bps_config_value(self.site, ".xlarge.scheduler_options", list, []),
            },
        }

    def _get_partition_for_slot(self, slot: str) -> str:
        """Return the Slurm partition Parsl must use to submit jobs for the
        job slot `slot`. Values of `slot` can be "small", "medium", "large"
        or "xlarge".
        """
        # The target Slurm partition must be selected according to the type of
        # the job slot but also according to the CPU architecture of the
        # compute node.
        #
        # Parsl requires that the CPU architecture of its orchestrator to
        # be identical to the architecture of its executors. Therefore,
        # we need to ensure that Slurm schedules our Parsl executors on
        # compute nodes with the same architecture as the host where this
        # orchestrator runs.

        # Default target Slurm partitions per CPU architecture
        default_partition = {
            "aarch64": {
                "small": "htc_arm",
                "medium": "htc_arm",
                "large": "htc_arm",
                "xlarge": "htc_arm",
            },
            "x86_64": {
                "small": "lsst,htc",
                "medium": "lsst",
                "large": "lsst",
                "xlarge": "lsst",
            },
        }
        architecture = platform.machine()
        if architecture not in default_partition:
            raise ValueError(f"architecture {architecture} is not supported")

        # If a partition was specified in the workflow description file
        # specifically for this job slot, use that partition. For instance:
        #
        # site:
        #   ccin2p3:
        #     class: lsst.ctrl.bps.parsl.sites.ccin2p3.Ccin2p3
        #     small:
        #       partition: htc
        slot_partition = get_bps_config_value(self.site, f".{slot}.partition", str, "")
        if slot_partition != "":
            return slot_partition

        # If a partition was specified in the workflow description file at
        # the site level, use that partition. For instance:
        #
        # site:
        #   ccin2p3:
        #     class: lsst.ctrl.bps.parsl.sites.ccin2p3.Ccin2p3
        #     partition: htc
        #
        # Otherwise, use the default for this slot on this architecture.
        return get_bps_config_value(self.site, ".partition", str, default_partition[architecture][slot])

    def _get_walltime_for_slot(self, slot: str) -> str:
        """Return the value for walltime Parsl must use to submit jobs for the
        job slot `slot`. Values of `slot` can be "small", "medium", "large"
        or "xlarge".
        """
        # If a specific walltime value was specified for this job slot in the
        # configuration use that value. For instance:
        #
        # site:
        #   ccin2p3:
        #     class: lsst.ctrl.bps.parsl.sites.ccin2p3.Ccin2p3
        #     small:
        #       walltime: "3:00:00"
        slot_walltime = get_bps_config_value(self.site, f".{slot}.walltime", str, "")
        if slot_walltime != "":
            return slot_walltime

        # If a walltime value was specified for the site use that value.
        # Otherwise, use the default walltime. For instance:
        #
        # site:
        #   ccin2p3:
        #     class: lsst.ctrl.bps.parsl.sites.ccin2p3.Ccin2p3
        #     walltime: "3:00:00"
        return get_bps_config_value(self.site, ".walltime", str, self.DEFAULT_WALLTIME)

    def get_executors(self) -> list[ParslExecutor]:
        """Get a list of Parsl executors that can be used for processing a
        workflow.

        Each executor must have a unique ``label``.
        """
        executors: list[ParslExecutor] = []
        for label, slot in self._slot_size.items():
            # Compute the scheduler options for this job slot. Options
            # specified at the slot level in the configuration file
            # overwrite those specified at the site level.
            scheduler_options = copy.deepcopy(self._scheduler_options)
            if slot_scheduler_options := slot.get("scheduler_options", []):
                scheduler_options = copy.deepcopy(slot_scheduler_options)

            options = f"#SBATCH {' '.join(opt for opt in scheduler_options)}" if scheduler_options else ""

            executor = HighThroughputExecutor(
                label,
                provider=SlurmProvider(
                    # Slurm partition to request blocks from.
                    partition=slot["partition"],
                    # Slurm account to which to charge resources used by the
                    # job.
                    account=self._account,
                    # Nodes to provision per block (1 block = 1 CPU core).
                    nodes_per_block=1,
                    # Number of CPU cores to provision per node.
                    cores_per_node=1,
                    # Memory per node (GB) for each Slurm job.
                    mem_per_node=slot["memory"],
                    # Initial number of blocks.
                    init_blocks=0,
                    # Minimum number of blocks to maintain.
                    min_blocks=0,
                    # Maximum number of blocks to maintain.
                    max_blocks=slot["max_blocks"],
                    # Time limit for each Slurm job.
                    walltime=slot["walltime"],
                    # '#SBATCH' directives to prepend to the Slurm submission
                    # script.
                    scheduler_options=options,
                    # Set the number of file descriptors and processes to
                    # the maximum allowed.
                    worker_init="ulimit -n hard && ulimit -u hard",
                    # Requests nodes which are not shared with other running
                    # jobs.
                    exclusive=False,
                ),
                # Address to connect to the main Parsl process.
                address=self.get_address(),
                # GB of memory required per worker. If specified the node
                # manager will check the available memory at startup and limit
                # the number of workers such that the there’s sufficient memory
                # for each worker.
                mem_per_worker=None,
                # Caps the number of workers launched per node.
                max_workers_per_node=1,
                # Timeout period (in milliseconds) to be used by the
                # executor components.
                poll_period=1_000,
                # Retry submitting to Slurm in case of submission error.
                block_error_handler=False,
            )
            executors.append(executor)

        return executors

    def select_executor(self, job: "ParslJob") -> str:
        """Get the ``label`` of the executor to use to execute ``job``.

        Parameters
        ----------
        job : `ParslJob`
            Job to be executed.

        Returns
        -------
        label : `str`
            Label of executor to use to execute ``job``.
        """
        # We choose the executor to use based only on the memory required
        # by the job.
        memory = job.generic.request_memory / 1024  # Convert to GB
        for label in ("small", "medium", "large"):
            if memory <= self._slot_size[label]["memory"]:
                return label

        return "xlarge"

    def get_parsl_config(self) -> parsl.config.Config:
        """Get Parsl configuration for using CC-IN2P3 Slurm farm as a
        Parsl execution site.

        Returns
        -------
        config : `parsl.config.Config`
            The configuration to be used to initialize Parsl for this site.
        """
        executors = self.get_executors()
        monitor = self.get_monitor()

        # Number of retries in case of job failure.
        retries = get_bps_config_value(self.site, ".retries", int, 0)

        # Path to run directory.
        run_dir = get_bps_config_value(self.site, ".run_dir", str, "parsl_runinfo")

        # Strategy for scaling blocks according to workflow needs.
        # Use a strategy that allows for scaling up and down Parsl workers.
        strategy = get_bps_config_value(self.site, ".strategy", str, "htex_auto_scale")

        return parsl.config.Config(
            executors=executors,
            monitoring=monitor,
            retries=retries,
            checkpoint_mode="task_exit",
            run_dir=run_dir,
            strategy=strategy,
        )



# This file is part of ctrl_bps_parsl.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org) and the LSST DESC (https://www.lsstdesc.org/).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Any

from parsl.executors import HighThroughputExecutor
from parsl.executors.base import ParslExecutor
from parsl.providers import SlurmProvider

from ..configuration import get_bps_config_value, get_workflow_name
from ..site import SiteConfig

if TYPE_CHECKING:
    from ..job import ParslJob

__all__ = ("Slurm", "TripleSlurm")


Kwargs = dict[str, Any]


class Slurm(SiteConfig):
    """Configuration for generic Slurm cluster.

    This can be used directly as the site configuration for a Slurm cluster by
    setting the BPS config, e.g.:

    .. code-block:: yaml

        computeSite: slurm
        site:
          slurm:
            class: lsst.ctrl.bps.parsl.sites.Slurm
            nodes: 3
            cores_per_node: 20
            walltime: "00:59:00"  # Note: always quote walltime in YAML

    Alternatively, it can be used as a base class for Slurm cluster
    configurations.

    The following BPS configuration parameters are recognised (and required
    unless there is a default mentioned here, or provided by a subclass):

    - ``nodes`` (`int`): number of nodes for each Slurm job.
    - ``cores_per_node`` (`int`): number of cores per node for each Slurm job;
      by default we use all cores on the node.
    - ``walltime`` (`str`): time limit for each Slurm job.
    - ``mem_per_node`` (`int`): memory per node (GB) for each Slurm job; by
      default we use whatever Slurm gives us.
    - ``qos`` (`str`): quality of service to request for each Slurm job; by
      default we use whatever Slurm gives us.
    - ``singleton`` (`bool`): allow only one job to run at a time; by default
      ``False``.
    - ``account`` (`str`): account to use for Slurm jobs.
    - ``scheduler_options`` (`str`): text to prepend to the Slurm submission
      script (each line usually starting with ``#SBATCH``).
    """

    def make_executor(
        self,
        label: str,
        *,
        nodes: int | None = None,
        cores_per_node: int | None = None,
        walltime: str | None = None,
        mem_per_node: int | None = None,
        mem_per_worker: float | None = None,
        qos: str | None = None,
        constraint: str | None = None,
        singleton: bool = False,
        scheduler_options: str | None = None,
        provider_options: Kwargs | None = None,
        executor_options: Kwargs | None = None,
    ) -> ParslExecutor:
        """Return an executor for running on a Slurm cluster.

        Parameters
        ----------
        label : `str`
            Label for executor.
        nodes : `int`, optional
            Default number of nodes for each Slurm job.
        cores_per_node : `int`, optional
            Default number of cores per node for each Slurm job.
        walltime : `str`, optional
            Default time limit for each Slurm job.
        mem_per_node : `float`, optional
            Memory per node (GB) to request for each Slurm job.
        mem_per_worker : `float`, optional
            Minimum memory per worker (GB), limited by the executor.
        qos : `str`, optional
            Quality of service for each Slurm job.
        constraint : `str`, optional
            Node feature(s) to require for each Slurm job.
        singleton : `bool`, optional
            Wether to allow only a single Slurm job to run at a time.
        scheduler_options : `str`, optional
            ``#SBATCH`` directives to prepend to the Slurm submission script.
        provider_options : `dict`, optional
            Additional arguments for `SlurmProvider` constructor.
        executor_options : `dict`, optional
            Additional arguments for `HighThroughputExecutor` constructor.

        Returns
        -------
        executor : `HighThroughputExecutor`
            Executor for Slurm jobs.
        """
        nodes = get_bps_config_value(self.site, "nodes", int, nodes, required=True)
        cores_per_node = get_bps_config_value(self.site, "cores_per_node", int, cores_per_node)
        walltime = get_bps_config_value(self.site, "walltime", str, walltime, required=True)
        mem_per_node = get_bps_config_value(self.site, "mem_per_node", int, mem_per_node)
        qos = get_bps_config_value(self.site, "qos", str, qos)
        singleton = get_bps_config_value(self.site, "singleton", bool, singleton)
        account = get_bps_config_value(self.site, "account", str)
        scheduler_options = get_bps_config_value(self.site, "scheduler_options", str, scheduler_options)

        job_name = get_workflow_name(self.config)
        if scheduler_options is None:
            scheduler_options = ""
        else:
            scheduler_options += "\n"
        scheduler_options += f"#SBATCH --job-name={job_name}\n"
        if qos:
            scheduler_options += f"#SBATCH --qos={qos}\n"
        if constraint:
            scheduler_options += f"#SBATCH --constraint={constraint}\n"
        if singleton:
            # The following SBATCH directives allow only a single slurm job
            # (parsl block) with our job_name to run at once. This means we can
            # have one job running, and one already in the queue when the first
            # exceeds the walltime limit. More backups could be achieved with a
            # larger value of max_blocks. This only allows one job to be
            # actively running at once, so that needs to be sized appropriately
            # by the user.
            scheduler_options += "#SBATCH --dependency=singleton\n"
        return HighThroughputExecutor(
            label,
            provider=SlurmProvider(
                nodes_per_block=nodes,
                cores_per_node=cores_per_node,
                mem_per_node=mem_per_node,
                walltime=walltime,
                account=account,
                scheduler_options=scheduler_options,
                **(provider_options or {}),
            ),
            mem_per_worker=mem_per_worker,
            address=self.get_address(),
            **(executor_options or {}),
        )

    def get_executors(self) -> list[ParslExecutor]:
        """Get a list of executors to be used in processing.

        Each executor should have a unique ``label``.
        """
        return [self.make_executor("slurm")]

    def select_executor(self, job: "ParslJob") -> str:
        """Get the ``label`` of the executor to use to execute a job.

        Parameters
        ----------
        job : `ParslJob`
            Job to be executed.

        Returns
        -------
        label : `str`
            Label of executor to use to execute ``job``.
        """
        return "slurm"


class TripleSlurm(Slurm):
    """Configuration for running jobs on a Slurm cluster with three levels.

    Parameters
    ----------
    *args : `~typing.Any`
        Parameters forwarded to base class constructor.
    **kwargs : `~typing.Any`
        Keyword arguments passed to base class constructor.

    Notes
    -----
    The three levels are useful for having workers with different amount of
    available memory (and this is how executors are selected, by default),
    though other uses are possible.

    The following BPS configuration parameters are recognised, overriding the
    defaults:

    - ``nodes`` (`int`): number of nodes for each Slurm job.
    - ``cores_per_node`` (`int`): number of cores per node for each Slurm job;
      by default we use all cores on the node.
    - ``walltime`` (`str`): time limit for each Slurm job; setting this would
      override each of the ``small_walltime``, ``medium_walltime`` and
      ``large_walltime`` values.
    - ``mem_per_node`` (`float`): memory per node for each Slurm job; by
      default we use whatever Slurm gives us.
    - ``qos`` (`str`): quality of service to request for each Slurm job; by
      default we use whatever Slurm gives us.
    - ``small_memory`` (`float`): memory per worker (GB) for each 'small' Slurm
      job.
    - ``medium_memory`` (`float`): memory per worker (GB) for each 'medium'
      Slurm job.
    - ``large_memory`` (`float`): memory per worker (GB) for each 'large' Slurm
      job.
    - ``small_walltime`` (`str`): time limit for each 'small' Slurm job.
    - ``medium_walltime`` (`str`): time limit for each 'medium' Slurm job.
    - ``large_walltime`` (`str`): time limit for each 'large' Slurm job.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.small_memory = get_bps_config_value(self.site, "small_memory", float, 2.0)
        self.medium_memory = get_bps_config_value(self.site, "medium_memory", float, 4.0)
        self.large_memory = get_bps_config_value(self.site, "large_memory", float, 8.0)
        self.small_walltime = get_bps_config_value(self.site, "small_walltime", str, "10:00:00")
        self.medium_walltime = get_bps_config_value(self.site, "medium_walltime", str, "10:00:00")
        self.large_walltime = get_bps_config_value(self.site, "large_walltime", str, "40:00:00")

    def get_executors(
        self,
        small_options: Kwargs | None = None,
        medium_options: Kwargs | None = None,
        large_options: Kwargs | None = None,
        **common_options,
    ) -> list[ParslExecutor]:
        """Get a list of executors to be used in processing.

        We create three executors, with different walltime and memory per
        worker.

        Parameters
        ----------
        small_options : kwargs
            Options for ``make_executor`` for small executor.
        medium_options : kwargs
            Options for ``make_executor`` for medium executor.
        large_options : kwargs
            Options for ``make_executor`` for large executor.
        **common_options
            Common options for ``make_executor`` for each of the executors.
        """
        if small_options is None:
            small_options = {}
        if medium_options is None:
            medium_options = {}
        if large_options is None:
            large_options = {}

        small_options["walltime"] = small_options.get("walltime", self.small_walltime)
        medium_options["walltime"] = medium_options.get("walltime", self.medium_walltime)
        large_options["walltime"] = large_options.get("walltime", self.large_walltime)

        small_options["mem_per_worker"] = small_options.get("mem_per_worker", self.small_memory)
        medium_options["mem_per_worker"] = medium_options.get("mem_per_worker", self.medium_memory)
        large_options["mem_per_worker"] = large_options.get("mem_per_worker", self.large_memory)

        return [
            self.make_executor("small", **small_options, **(common_options or {})),
            self.make_executor("medium", **medium_options, **(common_options or {})),
            self.make_executor("large", **large_options, **(common_options or {})),
        ]

    def select_executor(self, job: "ParslJob") -> str:
        """Get the ``label`` of the executor to use to execute a job.

        This implementation only looks at the requested memory.

        Parameters
        ----------
        job : `ParslJob`
            Job to be executed.

        Returns
        -------
        label : `str`
            Label of executor to use to execute ``job``.
        """
        memory = job.generic.request_memory / 1024  # GB
        if memory <= self.small_memory:
            return "small"
        if memory <= self.medium_memory:
            return "medium"
        return "large"



# This file is part of ctrl_bps_parsl.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org) and the LSST DESC (https://www.lsstdesc.org/).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
import subprocess
from collections import defaultdict
from collections.abc import Sequence
from functools import partial
from textwrap import dedent
from typing import Any

from parsl.app.bash import BashApp
from parsl.app.futures import Future

from lsst.ctrl.bps import BpsConfig, GenericWorkflow, GenericWorkflowJob

from .configuration import get_bps_config_value

__all__ = ("ParslJob", "get_file_paths")

_env_regex = re.compile(r"<ENV:(\S+)>")  # Regex for replacing <ENV:WHATEVER> in BPS job command-lines
_file_regex = re.compile(r"<FILE:(\S+)>")  # Regex for replacing <FILE:WHATEVER> in BPS job command-lines


def run_command(
    command_line: str,
    inputs: Sequence[Future] = (),
    stdout: str | None = None,
    stderr: str | None = None,
    parsl_resource_specification: dict[str, Any] | None = None,
) -> str:
    """Run a command.

    This function exists to get information into parsl, through the ``inputs``,
    ``stdout`` and ``stderr`` parameters. It needs to be wrapped by a parsl
    ``bash_app`` decorator before use, after which it will return a `Future`.

    Parameters
    ----------
    command_line : `str`
        Command-line to have parsl run.
    inputs : list of `Future`
        Other commands that must have run before this.
    stdout, stderr : `str`, optional
        Filenames for stdout and stderr.
    parsl_resource_specification : `dict`, optional
        Resources required for job.

    Returns
    -------
    command_line : `str`
        Command-line to have parsl run.
    """
    return command_line


def get_file_paths(workflow: GenericWorkflow, name: str) -> dict[str, str]:
    """Extract file paths for a job.

    Parameters
    ----------
    workflow : `GenericWorkflow`
        BPS workflow that knows the file paths.
    name : `str`
        Job name.

    Returns
    -------
    paths : `dict` mapping `str` to `str`
        File paths for job, indexed by symbolic name.
    """
    return {ff.name: ff.src_uri for ff in workflow.get_job_inputs(name)}


class ParslJob:
    """Job to execute with parsl.

    Parameters
    ----------
    generic : `GenericWorkflowJob`
        BPS job information.
    config : `BpsConfig`
        BPS configuration.
    file_paths : `dict` mapping `str` to `str`
        File paths for job, indexed by symbolic name.
    """

    def __init__(
        self,
        generic: GenericWorkflowJob,
        config: BpsConfig,
        file_paths: dict[str, str],
    ):
        self.generic = generic
        self.name = generic.name
        self.config = config
        self.file_paths = file_paths
        self.future = None
        self.done = False

        # Determine directory for job stdout and stderr
        log_dir = os.path.join(get_bps_config_value(self.config, "submitPath", str, required=True), "logs")
        _, template = self.config.search(
            "subDirTemplate",
            opt={
                "curvals": {"curr_site": self.config["computeSite"], "curr_cluster": self.generic.label},
                "replaceVars": False,
                "default": "",
            },
        )
        job_vals = defaultdict(str)
        job_vals["label"] = self.generic.label
        if self.generic.tags:
            job_vals.update(self.generic.tags)
        subdir = template.format_map(job_vals)
        # Call normpath just to make paths easier to read as templates tend
        # to have variables that aren't used by every job.  Avoid calling on
        # empty string because changes it to dot.
        same_part = os.path.normpath(os.path.join(log_dir, subdir, self.name))
        self.stdout = same_part + ".stdout"
        self.stderr = same_part + ".stderr"

    def __reduce__(self):
        """Recipe for pickling."""
        return type(self), (self.generic, self.config, self.file_paths)

    def get_command_line(self, allow_stage=True) -> str:
        """Get the bash command-line to run to execute this job.

        Parameters
        ----------
        allow_stage : `bool`
            Allow staging of execution butler? This is not appropriate for the
            initial or final jobs that run on the local nodes.

        Returns
        -------
        command : `str`
            Command-line to execute for job.
        """
        command: str = self.generic.executable.src_uri + " " + self.generic.arguments
        if not allow_stage:
            return command
        exec_butler_dir = get_bps_config_value(self.config, "executionButlerDir", str)
        if not exec_butler_dir or not os.path.isdir(exec_butler_dir):
            # We're not using the execution butler
            return command

        # Add commands to copy the execution butler.
        # This keeps workers from overloading the sqlite database.
        # The copy can be deleted once we're done, because the original
        # execution butler contains everything that's required.
        job_dir = os.path.join(os.path.dirname(exec_butler_dir), self.name)
        # Set the butlerConfig field to the location of the job-specific copy.
        command = command.replace("<FILE:butlerConfig>", job_dir)
        return dedent(
            f"""
            if [[ ! -d {job_dir} ]]; then mkdir -p {job_dir}; fi
            cp {exec_butler_dir}/* {job_dir}
            {command}
            retcode=$?
            rm -rf {job_dir}
            exit $retcode
            """
        )

    def evaluate_command_line(self, command: str) -> str:
        """Evaluate the bash command-line.

        BPS provides a command-line with symbolic names for BPS variables,
        environment variables and files. Here, we replace those symbolic names
        with the actual values, to provide a concrete command that can be
        executed.

        In replacing file paths, we are implicitly assuming that we are working
        on a shared file system, i.e., that workers can see the butler
        directory, and that files do not need to be staged to the worker.

        Parameters
        ----------
        command : `str`
            Command-line to execute, from BPS.

        Returns
        -------
        command : `str`
            Command ready for execution on a worker.
        """
        command = command.format(**self.generic.cmdvals)  # BPS variables

        # Make sure *all* symbolic names are resolved.
        #
        # In general, actual values for some symbolic names may contain other
        # symbolic names. As a result, more than one iteration may be required
        # to resolve all symbolic names.  For example, an actual value for
        # a filename may contain a symbolic name for an environment variable.
        prev_command = command
        while True:
            command = re.sub(_env_regex, r"${\g<1>}", command)  # Environment variables
            command = re.sub(_file_regex, lambda match: self.file_paths[match.group(1)], command)  # Files
            if prev_command == command:
                break
            prev_command = command

        return command

    def get_resources(self) -> dict[str, Any]:
        """Return what resources are required for executing this job."""
        resources = {}
        for bps_name, parsl_name, scale in (
            ("request_memory", "memory", None),  # Both BPS and WorkQueueExecutor use MB
            ("request_cpus", "cores", None),
            ("request_disk", "disk", None),  # Both are MB
            ("request_walltime", "running_time_min", None),  # Both are minutes
        ):
            value = getattr(self.generic, bps_name)
            if scale is not None:
                value *= scale
            resources[parsl_name] = value
        return resources

    def get_future(
        self,
        app: BashApp,
        inputs: list[Future],
        command_prefix: str | None = None,
        add_resources: bool = False,
    ) -> Future | None:
        """Get the parsl app future for the job.

        This effectively queues the job for execution by a worker, subject to
        dependencies.

        Parameters
        ----------
        app : callable
            A parsl bash_app decorator to use.
        inputs : list of `Future`
            Dependencies to be satisfied before executing this job.
        command_prefix : `str`, optional
            Bash commands to execute before the job command, e.g., for setting
            the environment.
        add_resources : `bool`
            Add resource specification when submitting the job? This is only
            appropriate for the ``WorkQueue`` executor; other executors will
            raise an exception.

        Returns
        -------
        future : `Future` or `None`
            A `Future` object linked to the execution of the job, or `None` if
            the job has already been done (e.g., by ``run_local``).
        """
        if self.done:
            return None  # Nothing to do
        if not self.future:
            command = self.get_command_line()
            command = self.evaluate_command_line(command)
            if command_prefix:
                command = command_prefix + "\n" + command
            resources = self.get_resources() if add_resources else {}

            # Add a layer of indirection to which we can add a useful name.
            # This name is used by parsl for tracking workflow status.
            func = partial(run_command)
            func.__name__ = self.generic.label  # type: ignore

            self.future = app(func)(
                command,
                inputs=inputs,
                stdout=self.stdout,
                stderr=self.stderr,
                parsl_resource_specification=resources,
            )
        return self.future

    def run_local(self):
        """Run the command locally.

        This is intended to support jobs that should not be done by a
        worker.
        """
        if self.done:  # Nothing to do
            return
        command = self.get_command_line(False)
        command = self.evaluate_command_line(command)
        os.makedirs(os.path.dirname(self.stdout), exist_ok=True)
        os.makedirs(os.path.dirname(self.stderr), exist_ok=True)
        with open(self.stdout, "w") as stdout, open(self.stderr, "w") as stderr:
            subprocess.check_call(command, shell=True, executable="/bin/bash", stdout=stdout, stderr=stderr)
        self.done = True



# This file is part of ctrl_bps_parsl.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org) and the LSST DESC (https://www.lsstdesc.org/).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

__all__ = ("export_environment",)


def export_environment():
    """Generate bash script to regenerate the current environment."""
    output = ""
    for key, val in os.environ.items():
        if key in ("DISPLAY",):
            continue
        if val.startswith("() {"):
            # This is a function.
            # "Two parentheses, a single space, and a brace"
            # is exactly the same criterion as bash uses.

            # From 2014-09-25, the function name is prefixed by 'BASH_FUNC_'
            # and suffixed by '()', which we have to remove.
            if key.startswith("BASH_FUNC_") and key.endswith("()"):
                key = key[10:-2]

            output += f"{key} {val}\nexport -f {key}\n"
        else:
            # This is a variable.
            output += "export {key}='{val}'\n".format(key=key, val=val.replace("'", "'\"'\"'"))
    return output



# This file is part of ctrl_bps_parsl.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org) and the LSST DESC (https://www.lsstdesc.org/).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

from lsst.ctrl.bps import BpsConfig, GenericWorkflowJob
from lsst.ctrl.bps.parsl.job import ParslJob

TESTDIR = os.path.abspath(os.path.dirname(__file__))


def testInitWithTemplate():
    """Test ParslJob stdout/stderr init. with template."""
    # Test doesn't actually use directory
    submit_path = os.path.join(TESTDIR, "with_template")
    config = BpsConfig(
        {
            "subDirTemplate": "{label}/{tract}/{patch}/{band}/{visit}/{exposure}",
            "submitPath": submit_path,
        },
        search_order=[],
    )
    gwjob = GenericWorkflowJob("job1", label="label1")
    gwjob.tags = {"exposure": 903344}

    parsl_job = ParslJob(gwjob, config, {})
    assert parsl_job.stdout == os.path.join(submit_path, "logs/label1/903344/job1.stdout")
    assert parsl_job.stderr == os.path.join(submit_path, "logs/label1/903344/job1.stderr")


def testInitWithTemplateNoTags():
    """Test ParslJob stdout/stderr init. with template but without tags."""
    # Test doesn't actually use directory
    submit_path = os.path.join(TESTDIR, "with_template_no_tags")
    config = BpsConfig(
        {
            "subDirTemplate": "{label}/{tract}/{patch}/{band}/{visit}/{exposure}",
            "submitPath": submit_path,
        },
        search_order=[],
    )
    gwjob = GenericWorkflowJob("job2", label="label2")
    gwjob.tags = None

    parsl_job = ParslJob(gwjob, config, {})
    assert parsl_job.stdout == os.path.join(submit_path, "logs/label2/job2.stdout")
    assert parsl_job.stderr == os.path.join(submit_path, "logs/label2/job2.stderr")


def testInitWithoutTemplate():
    """Test ParslJob stdout/stderr init. with template."""
    # Test doesn't actually use directory
    submit_path = os.path.join(TESTDIR, "without_template")
    config = BpsConfig({"subDirTemplate": "", "submitPath": submit_path}, search_order=[])
    gwjob = GenericWorkflowJob("job3", label="label3")
    gwjob.tags = {"exposure": 903344}

    parsl_job = ParslJob(gwjob, config, {})
    assert parsl_job.stdout == os.path.join(submit_path, "logs/job3.stdout")
    assert parsl_job.stderr == os.path.join(submit_path, "logs/job3.stderr")



# This file is part of ctrl_bps_parsl.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org) and the LSST DESC (https://www.lsstdesc.org/).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING

from parsl.executors import HighThroughputExecutor
from parsl.executors.base import ParslExecutor
from parsl.providers import LocalProvider

from ..configuration import get_bps_config_value
from ..site import SiteConfig

if TYPE_CHECKING:
    from ..job import ParslJob

__all__ = ("Local",)


class Local(SiteConfig):
    """Configuration for running jobs on the local machine.

    The number of cores to use is specified in the site configuration, under
    ``site.<computeSite>.cores`` (`int`).
    """

    def get_executors(self) -> list[ParslExecutor]:
        """Get a list of executors to be used in processing.

        Each executor should have a unique ``label``.
        """
        cores = get_bps_config_value(self.site, "cores", int, required=True)
        return [HighThroughputExecutor("local", provider=LocalProvider(), max_workers_per_node=cores)]

    def select_executor(self, job: "ParslJob") -> str:
        """Get the ``label`` of the executor to use to execute a job.

        Parameters
        ----------
        job : `ParslJob`
            Job to be executed.

        Returns
        -------
        label : `str`
            Label of executor to use to execute ``job``.
        """
        return "local"
